// Resilient Wearables: Designing decentralised guerrilla networks
// Studio Weise7 + tm2024 workshop 
// studio@weise7.org

// Include Libraries
#include <esp_now.h>
#include <esp_wifi.h>
#include <WiFi.h>
#include <Adafruit_NeoPixel.h>

#define PIN 7  // ESP32-C3 built-in RGB led
#define NUMPIXELS 1

Adafruit_NeoPixel pixels(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ800);

// Variables for test data
int int_value;

// MAC Address of responder - edit as required
uint8_t broadcastAddress[] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

// Define a data structure
typedef struct struct_message {
  char a[32];
  int b;
} struct_message;

// Create a structured object
struct_message myData;

int packet_counter = 0;
float rssi;

// Peer info
esp_now_peer_info_t peerInfo;

typedef struct {
  unsigned frame_ctrl:16;
  unsigned duration_id:16;
  uint8_t addr1[6]; /* receiver address */
  uint8_t addr2[6]; /* sender address */
  uint8_t addr3[6]; /* filtering address */
  unsigned sequence_ctrl:16;
  uint8_t addr4[6]; /* optional */
} wifi_ieee80211_mac_hdr_t;

typedef struct {
  wifi_ieee80211_mac_hdr_t hdr;
  uint8_t payload[0]; /* network data ended with 4 bytes csum (CRC32) */
} wifi_ieee80211_packet_t;

void promiscuous_rx_cb(void *buf, wifi_promiscuous_pkt_type_t type) {

    // All espnow traffic uses action frames which are a subtype of the mgmnt frames so filter out everything else.
    if (type != WIFI_PKT_MGMT)
        return;

    static const uint8_t ACTION_SUBTYPE = 0xd0;
    static const uint8_t ESPRESSIF_OUI1[] = {0xA0, 0x76, 0x4E};
    static const uint8_t ESPRESSIF_OUI2[] = {0xC0, 0x4E, 0x30};

    const wifi_promiscuous_pkt_t *ppkt = (wifi_promiscuous_pkt_t *)buf;
    const wifi_ieee80211_packet_t *ipkt = (wifi_ieee80211_packet_t *)ppkt->payload;
    const wifi_ieee80211_mac_hdr_t *hdr = &ipkt->hdr;

    // Only continue processing if this is an action frame containing the Espressif OUI.
    if ((ACTION_SUBTYPE == (hdr->frame_ctrl & 0xFF)) && 
        (memcmp(hdr->addr2, ESPRESSIF_OUI1, 3) == 0) ||
        (memcmp(hdr->addr2, ESPRESSIF_OUI2, 3) == 0)) {
         rssi = ppkt->rx_ctrl.rssi;
         /*
         uint8_t addr1 = hdr->addr1[0];
         uint8_t addr2 = hdr->addr2[0];
         uint8_t addr3 = hdr->addr3[0];
         uint8_t addr4 = hdr->addr4[0];
         Serial.println(rssi);
         Serial.println(hdr->addr1[5]);
         Serial.println(hdr->addr2[5]);
         Serial.println(addr3);
         Serial.println(addr4);
         */
        packet_counter = 0;
    } else {
      if ( packet_counter < 1000 ) {
        packet_counter++;
      }
    }
    //Serial.println(packet_counter);
}

// Callback function called when data is sent
void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status) {
  Serial.print("\r\nLast Packet Send Status:\t");
  Serial.println(status == ESP_NOW_SEND_SUCCESS ? "Delivery Success" : "Delivery Fail");
}

// Callback function executed when data is received
void OnDataRecv(const uint8_t * mac, const uint8_t *incomingData, int len) {
  memcpy(&myData, incomingData, sizeof(myData));
  Serial.print("Data received: ");
  Serial.println(len);
  //Serial.print("Character Value: ");
  //Serial.println(myData.a);
  //Serial.print("Integer Value: ");
  //Serial.println(myData.b);
  //Serial.print("Float Value: ");
  //Serial.println(myData.c);
  //Serial.print("Boolean Value: ");
  //Serial.println(myData.d);
  //Serial.println();
}

void setup() {
  
  // Set up Serial Monitor
  Serial.begin(115200); 

  // Set ESP32 as a Wi-Fi Station
  WiFi.mode(WIFI_STA);

  pixels.begin();

  WiFi.setSleep(false);

  // Initilize ESP-NOW
  if (esp_now_init() != ESP_OK) {
    Serial.println("Error initializing ESP-NOW");
    return;
  }

  // Register the send callback
  esp_now_register_send_cb(OnDataSent);
  esp_wifi_set_promiscuous(true);
  esp_wifi_set_promiscuous_rx_cb(&promiscuous_rx_cb);
  
  // Register peer
  memcpy(peerInfo.peer_addr, broadcastAddress, 6);
  peerInfo.channel = 0;  
  peerInfo.encrypt = false;
  
  // Add peer        
  if (esp_now_add_peer(&peerInfo) != ESP_OK){
    //Serial.println("Failed to add peer");
    return;
  }
}

// mapping input to output
int output_start = 0;
int output_end = 255;
int input_start = 100;
int input_end = 0;

void loop() {

  int_value = random(1,20);

  // Format structured data
  strcpy(myData.a, "You are going amazing, sweety");
  myData.b = int_value;
  
  // Send message via ESP-NOW
  esp_err_t result = esp_now_send(broadcastAddress, (uint8_t *) &myData, sizeof(myData));

  int green = 0, red = 0, blue = 0; 
  
  if (result == ESP_OK) {
    //Serial.println("Sending confirmed");
    if ( packet_counter >= 1000 ) {
      rssi = -100;
    }
    int input = abs(rssi);
    if ( input > input_start ) {
      input = input_start;
    }

    double slope = 1.0 * (output_end - output_start) / (input_end - input_start );
    int prox = output_start + round(slope * (input - input_start));
    Serial.print("Proximity:\t\t\t");
    Serial.println(prox);

    switch (prox) {
      case 0 ... 29:
        green = 0, red = 0, blue = 255;
        break;
      case 49 ... 30:
        green = 0, red = 64, blue = 230;
        break;
      case 50 ... 80:
        green = 0, red = 180, blue = 180;
        break;
      case 81 ... 119:
        green = 0, red = 230, blue = 64;
        break;
      case 120 ... 255:
        green = 0, red = 255, blue = 0;
        break;
      default:
        break;
    }
  }
  else {
    Serial.println("Sending error");
    green = 0, red = 0, blue = 255;
  }

  pixels.clear();
  pixels.setPixelColor(0, pixels.Color(green, red, blue));
  pixels.show();

  delay(random(5,10)*100);
}